﻿using System;
using System.Linq;
using System.Collections.Generic;

namespace ConsoleApplication1
{
    class Person
    {
        public string Name { get; set; }
        public int Age { get; set; }
        public string Address { get; set; }

        public override string ToString()
        {
            return string.Format("{0}: {1} in {2}", Name, Age, Address);
        }
    }

    class MainLanguage
    {
        public string Name { get; set; }
        public string Language { get; set; }
    }

    class Program
    {
        static void Main(string[] args)
        {
            List<Person> people = new List<Person>
            {
                new Person { Name = "Tom", Age = 63, Address = "Korea" },
                new Person { Name = "Winnie", Age = 40, Address = "Tibet" },
                new Person { Name = "Anders", Age = 47, Address = "Sudan" },
                new Person { Name = "Hans", Age = 25, Address = "Tibet" },
                new Person { Name = "Eureka", Age = 32, Address = "Sudan" },
                new Person { Name = "Hawk", Age = 15, Address = "Korea" },
            };

            List<MainLanguage> languages = new List<MainLanguage>
            {
                new MainLanguage { Name = "Anders", Language = "Delphi" },
                new MainLanguage { Name = "Anders", Language = "C#" },
                new MainLanguage { Name = "Tom", Language = "Borland C++" },
                new MainLanguage { Name = "Hans", Language = "Visual C++" },
                new MainLanguage { Name = "Winnie", Language = "R" },
            };

            // LINQ 쿼리와 확장 메소드의 조합
            var all = from person in people
                      where person.Address == "Korea"
                      select person;

            var oldestAge = all.Max((elem) => elem.Age);
            Console.WriteLine(oldestAge);

            var oldestAge2 = people.Where((elem) => elem.Address == "Korea")
                    .Max((elem) => elem.Age);
            Console.WriteLine(oldestAge2);

            // LINQ 쿼리가 바로 실행됨
            Console.WriteLine("ToList() executed");
            var inKorea = (
                from person in people
                where IsEqual(person.Address, "Korea")
                select person
                            ).ToList();
            // 따라서, IsEqual 메소드의 실행으로 인해 화면에는 "Executed" 문자열이 출력됨
            Console.ReadLine();

            Console.WriteLine("IEnumerable<T> Where/Select evaluated");
            // IEnumerable<T>를 반환하므로 LINQ 쿼리가 평가만 되고 실행되지 않음
            var inKorea2 = from person in people
                           where IsEqual(person.Address, "Korea")
                           select person;

            // 따라서, IsEqual 메소드가 실행되지 않으므로 화면에는 "Executed" 문자열이 출력되지 않음.
            Console.ReadLine();

            // Take 확장 메소드 역시 IEnumerable<T>를 반환하므로 Executed 문자열이 출력되지 않음.
            Console.WriteLine("IEnumerable<T> Take evaluated");
            var firstPeople = inKorea2.Take(1);

            Console.ReadLine();

            // 열거를 시작했을 때 LINQ 쿼리가 실제로 실행됨.
            foreach (var item in firstPeople)
            {
                Console.WriteLine(item);
            }

            Console.ReadLine();

            // 또한 단일 값을 반환하는 Single 메소드의 호출은 곧바로 LINQ 쿼리가 실행되도록 만듦
            Console.WriteLine(firstPeople.Single());
        }

        private static bool IsEqual(string arg1, string arg2)
        {
            Console.WriteLine("Executed");
            return arg1 == arg2;
        }
    }
}
